#include <ddraw.h>
#include <ddrawex.h>

#include "prng.h"
#include "timer.h"
#include "math.h"

// 1D math

inline long DISTANCE(long a, long b)
{
	if (a < b)
		return (b - a);
	else
		return (a - b);											// Calculate magnitude
}

inline DWORD DISTANCE(DWORD a, DWORD b)
{
	if (a < b)
		return (b - a);
	else
		return (a - b);											// Calculate magnitude
}

inline long SGN(long a)
{
	if (a > 0)
		return 1;
	else if (a < 0)
		return -1;
	else
		return 0;
}

inline DWORD TRIM(DWORD a, DWORD b)
{
	if (a <= b)
		return (0);												// If the subtraction would produce a negative number (or 0 for speed), return 0.
	return (a - b);												// Else, do the subtraction
}

inline DWORD TRIMLOWER(DWORD a, DWORD max)
{
	if (a <= max)
		return a;
	return max;
}


// Mutex

class Mutex
{
	HANDLE hMutex;		// Mutex handle

public:
	Mutex()
	{
		hMutex = CreateMutex(NULL, false, NULL);
	}

	virtual ~Mutex()
	{
		CloseHandle(hMutex);
		hMutex = NULL;
	}

	void LockClass() const
	{
		WaitForSingleObject(hMutex, INFINITE);
	}

	void UnlockClass() const
	{
		ReleaseMutex(hMutex);
	}

	inline Mutex(const Mutex &copy)
	{
		hMutex = copy.hMutex;
	}
};


// Window settings

char * DXWindowClass = "SubSpaceClassOf2002";
char * DXWindowTitle = "SubSpace Client BUILD 1";


// Declarations

class DXWindow;

char *DDErrorString(HRESULT hr);


// Backprogression for subclassing

_linkedlist <DXWindow> DXWindowList;


class DXWindow
{
public:
	// Window information
	HWND hWnd;
	RECT SourceRect;
	LONG DestX, DestY;
	LONG Width, Height;
	bool GotFocus;
	bool Minimized;
	bool Open;

	// Window subclassing
	friend LRESULT CALLBACK WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		// Determine referenced window
		_listnode <DXWindow> * parse = DXWindowList.head;
		DXWindow * window = NULL;

		while (parse)
		{
			window = parse->item;

			if (window->hWnd == hWnd)
				break;

			parse = parse->next;
		}

		if (window == NULL)
		{
			MessageBox(NULL, "Unable to find referenced window.", "Subclassing error:", 0);

			return 0;
		}

		// Handle events
		switch (uMsg)
		{
		case WM_SIZE:
			{
				window->SourceRect.left = 0;
				window->SourceRect.top = 0;
				window->Width = window->SourceRect.right = LOWORD(lParam);
				window->HalfWidth = window->Width >> 1; 
				window->Height = window->SourceRect.bottom = HIWORD(lParam);
				window->HalfHeight = window->Height >> 1; 
			}
			break;
		case WM_SIZING:
			{
				LPRECT size = (LPRECT)lParam;

				if (size->right - size->left < 320)
					size->right = size->left + 320;
				if (size->bottom - size->top < 240)
					size->bottom = size->top + 240;
			}
			break;
		case WM_MOVE:
			{
				window->DestX = LOWORD(lParam);
				window->DestY = HIWORD(lParam);
			}
			break;
		case WM_ACTIVATE:
			{
				window->Minimized = (HIWORD(wParam) != 0); 
			}
			break;
		case WM_SHOWWINDOW:
			{
				window->GotFocus = (wParam != 0);
				window->Minimized = (lParam & SW_PARENTCLOSING);
			}
			break;
		case WM_KEYDOWN:
			switch (wParam)
			{
			case VK_LEFT:
				if (window->CenterX >= 50)
					window->CenterX -= 50;
				break;
			case VK_RIGHT:
				if (window->CenterX <= (1024 << 4) - 50)
					window->CenterX += 50;
				break;
			case VK_UP:
				if (window->CenterY >= 50)
					window->CenterY -= 50;
				break;
			case VK_DOWN:
				if (window->CenterY <= (1024 << 4) - 50)
					window->CenterY += 50;
				break;
			};
			break;
		case WM_QUIT:
		case WM_CLOSE:
			{
				window->Open = false;
			}
			break;
		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
		};

		return 0;
	}

	// DirectX surfaces
	IDirectDraw7 * pDD;
	LPDIRECTDRAWSURFACE7 Screen;
	LPDIRECTDRAWSURFACE7 Backbuffer;
	LPDIRECTDRAWCLIPPER Clipper;
	DDSURFACEDESC2 ddsd;
	DDBLTFX ddfx;

	DWORD CenterX, CenterY;

	long HalfWidth, HalfHeight;

	// Control
	DXWindow();
	~DXWindow();

	inline void Flip();
	inline void ClearSurface();
	inline void Translate(DWORD x, DWORD y);

	inline void Lock();
	inline void SetPixel(DWORD x, DWORD y, DWORD r, DWORD g, DWORD b);
	inline void SetPixel8(DWORD x, DWORD y, BYTE color);
	inline void Unlock();

	BYTE  * VideoMemory;
	DWORD	VideoPitch, VideoXShift, VideoIgnoreMask, VideoMask;
	DWORD	ShiftRed, ShiftGreen, ShiftBlue;
	DWORD	RedBitMask, GreenBitMask, BlueBitMask;
};


DXWindow::DXWindow()
{
#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "In DXWindow() constructor", "Loading...", 0);
#endif

	/* Initialize variables */
	Screen				= NULL;	// Screen
	Clipper				= NULL;	// Clipper
	Backbuffer			= NULL;	// Backbuffer
	pDD					= NULL;	// DirectDraw3
    IDirectDraw	  * IDD = NULL;	// DirectDraw1
	LONG retval;				// Function return value container
	memset(&SourceRect, 16, 0);	// Upper-lefthand corner of the screen
	hWnd				= 0;	// Window handle
	DestX				= 0;	// Window body X
	DestY				= 0;	// Window body Y
	Width				= 0;	// Window body width
	Height				= 0;	// Window body height
	Open				= true;	// Open?
	Minimized			= false;// Minimized?
	GotFocus			= true;	// Has focus?
	HalfWidth			= 0;	// Centralized calculation for _tiledsurface speed advantage
	HalfHeight			= 0;	// Centralized calculation for _tiledsurface speed advantage
	CenterX				= 8000;	// Camera X
	CenterY				= 8000;	// Camera Y
	ZeroMemory(&ddfx, sizeof(ddfx));
	ddfx.dwSize		= sizeof(ddfx);

	/* Register this DXWindow */
	DXWindowList.Append(this);

	/* Register the window class */
	WNDCLASS WndClass;
	WndClass.lpszClassName	= DXWindowClass;
	WndClass.cbClsExtra		= 0;
	WndClass.cbWndExtra		= 0;
 	WndClass.hIcon			= LoadIcon (NULL, IDI_APPLICATION);
	WndClass.hCursor		= LoadCursor (NULL, IDC_ARROW);
	WndClass.style			= NULL;
	WndClass.lpszMenuName	= NULL;
	WndClass.hInstance		= GetModuleHandle(NULL);
	WndClass.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
	WndClass.lpfnWndProc	= WindowProc;

	RegisterClass(&WndClass);

	/* Create the bot window */
    hWnd = CreateWindow(DXWindowClass,
						DXWindowTitle,
						WS_OVERLAPPEDWINDOW,
						CW_USEDEFAULT,
						CW_USEDEFAULT,
						CW_USEDEFAULT,
						CW_USEDEFAULT,
						0,
						0,
						GetModuleHandle(NULL),
						0);

	if (!hWnd)
	{
		_string S;
		S = "CreateWindow Error # ";
		S += (long) GetLastError();
		MessageBox(NULL, S.Text, "DXWindow::Initialize()", 0);
	}

	ShowWindow(hWnd, SW_SHOW);

	/* Produce a DirectDraw3 Interface */
	retval = DirectDrawCreateEx(NULL, (void**)&IDD, IID_IDirectDraw7, 0);

	if (!IDD)
	{
		_string S;
		S = "CreateDirectDraw Error # ";
		S += retval;
		S += ": ";
		S += DDErrorString(retval);
		MessageBox(NULL, S.Text, "DXWindow::Initialize()", 0);
	}

	retval = IDD->QueryInterface(IID_IDirectDraw7, (void**) &pDD);
	IDD->Release();

	if (!pDD)
	{
		_string S;
		S = "QueryInterface Error # ";
		S += retval;
		S += ": ";
		S += DDErrorString(retval);
		MessageBox(NULL, S.Text, "DXWindow::Initialize()", 0);
	}

	/* Set Windowed Mode */
    pDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL);

	/* Create the primary surface */
    ZeroMemory(&ddsd, sizeof(ddsd));
    ddsd.dwSize		= sizeof(ddsd);

	ddsd.dwFlags		= DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

    retval = pDD->CreateSurface(&ddsd, &Screen, NULL);

	if (!Screen)
	{
		_string S;
		S = "CreateSurface(Screen) Error # ";
		S += retval;
		S += ": ";
		S += DDErrorString(retval);
		MessageBox(NULL, S.Text, "DXWindow::Initialize()", 0);
	}

	/* Create the backbuffer surface */
    ZeroMemory(&ddsd, sizeof(ddsd));
    ddsd.dwSize		= sizeof(ddsd);

	ddsd.dwFlags		= DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth	= 1700;
	ddsd.dwHeight	= 1300;

	retval = pDD->CreateSurface(&ddsd, &Backbuffer, NULL);

	if (!Backbuffer)
	{
		_string S;
		S = "CreateSurface(Backbuffer) Error # ";
		S += retval;
		S += ": ";
		S += DDErrorString(retval);
		MessageBox(NULL, S.Text, "DXWindow::Initialize()", 0);
	}

	/* Create clipper */
	pDD->CreateClipper(0, &Clipper, NULL);
	Clipper->SetHWnd(0, hWnd);
	Screen->SetClipper(Clipper);

	/* Clear the surface */
	ClearSurface();

#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "Out of DXWindow() constructor", "Loading...", 0);
#endif
}


DXWindow::~DXWindow()
{
	/* Delete surfaces */
	if (Backbuffer)	Backbuffer->Release();
	if (Screen)		Screen->Release();
	if (pDD)		pDD->Release();
	if (Clipper)	Clipper->Release();

	/* Delete windows */
	DestroyWindow(hWnd);
	UnregisterClass(DXWindowClass, GetModuleHandle(NULL));

	/* Unregister this DXWindow */
	DXWindowList.Delete(this, true);
}


inline void GetMaskInfo(DWORD BitMask, DWORD &Shift)
{
	Shift = 0;

	do
	{
		if (BitMask & 1)
			return;

		Shift++;
	} while (BitMask >>= 1);
}


inline void DXWindow::Lock()
{
	if (Backbuffer->Lock(NULL, &ddsd, DDLOCK_NOSYSLOCK | DDLOCK_WRITEONLY | DDLOCK_WAIT, NULL) == DD_OK)
	{
		VideoPitch		= ddsd.lPitch;
		VideoMemory		= (BYTE*)ddsd.lpSurface;

		RedBitMask		= ddsd.ddpfPixelFormat.dwRBitMask;
		GreenBitMask	= ddsd.ddpfPixelFormat.dwGBitMask;
		BlueBitMask		= ddsd.ddpfPixelFormat.dwBBitMask;

		if (RedBitMask == 0)
		{
			RedBitMask		= 3;
			GreenBitMask	= 12;
			BlueBitMask		= 48;
		}

		GetMaskInfo(RedBitMask, ShiftRed);
		GetMaskInfo(GreenBitMask, ShiftGreen);
		GetMaskInfo(BlueBitMask, ShiftBlue);

		VideoIgnoreMask = (RedBitMask | GreenBitMask | BlueBitMask);
		VideoIgnoreMask = ~VideoIgnoreMask;

		VideoMask = VideoIgnoreMask & ((1 << ddsd.ddpfPixelFormat.dwRGBBitCount) - 1);
	}
	else
	{
		VideoMemory = NULL;
	}
}


inline void DXWindow::SetPixel(DWORD x, DWORD y, DWORD r, DWORD g, DWORD b)
{
	if (VideoMemory)
	{
		DWORD Offset = (x << 2) + y * VideoPitch;
		*(DWORD*)&VideoMemory[Offset] = VideoMask | ((r << ShiftRed) & RedBitMask) | ((g << ShiftGreen) & GreenBitMask) | ((b << ShiftBlue) & BlueBitMask) | (VideoIgnoreMask & *(DWORD*)&VideoMemory[Offset]);
	}
}


inline void DXWindow::SetPixel8(DWORD x, DWORD y, BYTE color)
{
	if (VideoMemory)
	{
		DWORD Offset = x + y * VideoPitch;
		VideoMemory[Offset] = color;
	}
}


inline void DXWindow::Unlock()
{
	Backbuffer->Unlock(NULL);
}


inline void DXWindow::Flip()
{
	RECT    rcDest;
	POINT   p;

	p.x = 0;
	p.y = 0;
	ClientToScreen(hWnd, &p);
	GetClientRect(hWnd, &rcDest);
	OffsetRect(&rcDest, p.x, p.y);

	HRESULT retval = Screen->Blt(&rcDest, Backbuffer, &SourceRect, DDBLT_WAIT, NULL);

	if ((retval != DD_OK) && (retval != DDERR_INVALIDRECT))
	{
		_string S;
		S = "Screen->Blt() Error # ";
		S += retval;
		S += ": ";
		S += DDErrorString(retval);
		MessageBox(NULL, S.Text, "DXWindow::Flip()", 0);
	}
}


inline void DXWindow::ClearSurface()
{
	Backbuffer->Blt(&SourceRect, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddfx);
}


inline void DXWindow::Translate(DWORD x, DWORD y)
{
	CenterX = x;
	CenterY = y;
}


char *DDErrorString(HRESULT hr)
{
	switch (hr)
	{
	case DDERR_ALREADYINITIALIZED:           return "DDERR_ALREADYINITIALIZED";
	case DDERR_CANNOTATTACHSURFACE:          return "DDERR_CANNOTATTACHSURFACE";
	case DDERR_CANNOTDETACHSURFACE:          return "DDERR_CANNOTDETACHSURFACE";
	case DDERR_CURRENTLYNOTAVAIL:            return "DDERR_CURRENTLYNOTAVAIL";
	case DDERR_EXCEPTION:                    return "DDERR_EXCEPTION";
	case DDERR_GENERIC:                      return "DDERR_GENERIC";
	case DDERR_HEIGHTALIGN:                  return "DDERR_HEIGHTALIGN";
	case DDERR_INCOMPATIBLEPRIMARY:          return "DDERR_INCOMPATIBLEPRIMARY";
	case DDERR_INVALIDCAPS:                  return "DDERR_INVALIDCAPS";
	case DDERR_INVALIDCLIPLIST:              return "DDERR_INVALIDCLIPLIST";
	case DDERR_INVALIDMODE:                  return "DDERR_INVALIDMODE";
	case DDERR_INVALIDOBJECT:                return "DDERR_INVALIDOBJECT";
	case DDERR_INVALIDPARAMS:                return "DDERR_INVALIDPARAMS";
	case DDERR_INVALIDPIXELFORMAT:           return "DDERR_INVALIDPIXELFORMAT";
	case DDERR_INVALIDRECT:                  return "DDERR_INVALIDRECT";
	case DDERR_LOCKEDSURFACES:               return "DDERR_LOCKEDSURFACES";
	case DDERR_NO3D:                         return "DDERR_NO3D";
	case DDERR_NOALPHAHW:                    return "DDERR_NOALPHAHW";
	case DDERR_NOCLIPLIST:                   return "DDERR_NOCLIPLIST";
	case DDERR_NOCOLORCONVHW:                return "DDERR_NOCOLORCONVHW";
	case DDERR_NOCOOPERATIVELEVELSET:        return "DDERR_NOCOOPERATIVELEVELSET";
	case DDERR_NOCOLORKEY:                   return "DDERR_NOCOLORKEY";
	case DDERR_NOCOLORKEYHW:                 return "DDERR_NOCOLORKEYHW";
	case DDERR_NODIRECTDRAWSUPPORT:          return "DDERR_NODIRECTDRAWSUPPORT";
	case DDERR_NOEXCLUSIVEMODE:              return "DDERR_NOEXCLUSIVEMODE";
	case DDERR_NOFLIPHW:                     return "DDERR_NOFLIPHW";
	case DDERR_NOGDI:                        return "DDERR_NOGDI";
	case DDERR_NOMIRRORHW:                   return "DDERR_NOMIRRORHW";
	case DDERR_NOTFOUND:                     return "DDERR_NOTFOUND";
	case DDERR_NOOVERLAYHW:                  return "DDERR_NOOVERLAYHW";
	case DDERR_NORASTEROPHW:                 return "DDERR_NORASTEROPHW";
	case DDERR_NOROTATIONHW:                 return "DDERR_NOROTATIONHW";
	case DDERR_NOSTRETCHHW:                  return "DDERR_NOSTRETCHHW";
	case DDERR_NOT4BITCOLOR:                 return "DDERR_NOT4BITCOLOR";
	case DDERR_NOT4BITCOLORINDEX:            return "DDERR_NOT4BITCOLORINDEX";
	case DDERR_NOT8BITCOLOR:                 return "DDERR_NOT8BITCOLOR";
	case DDERR_NOTEXTUREHW:                  return "DDERR_NOTEXTUREHW";
	case DDERR_NOVSYNCHW:                    return "DDERR_NOVSYNCHW";
	case DDERR_NOZBUFFERHW:                  return "DDERR_NOZBUFFERHW";
	case DDERR_NOZOVERLAYHW:                 return "DDERR_NOZOVERLAYHW";
	case DDERR_OUTOFCAPS:                    return "DDERR_OUTOFCAPS";
	case DDERR_OUTOFMEMORY:                  return "DDERR_OUTOFMEMORY";
	case DDERR_OUTOFVIDEOMEMORY:             return "DDERR_OUTOFVIDEOMEMORY";
	case DDERR_OVERLAYCANTCLIP:              return "DDERR_OVERLAYCANTCLIP";
	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE: return "DDERR_OVERLAYCOLORKEYONLYONEACTIVE";
	case DDERR_PALETTEBUSY:                  return "DDERR_PALETTEBUSY";
	case DDERR_COLORKEYNOTSET:               return "DDERR_COLORKEYNOTSET";
	case DDERR_SURFACEALREADYATTACHED:       return "DDERR_SURFACEALREADYATTACHED";
	case DDERR_SURFACEALREADYDEPENDENT:      return "DDERR_SURFACEALREADYDEPENDENT";
	case DDERR_SURFACEBUSY:                  return "DDERR_SURFACEBUSY";
	case DDERR_CANTLOCKSURFACE:              return "DDERR_CANTLOCKSURFACE";
	case DDERR_SURFACEISOBSCURED:            return "DDERR_SURFACEISOBSCURED";
	case DDERR_SURFACELOST:                  return "DDERR_SURFACELOST";
	case DDERR_SURFACENOTATTACHED:           return "DDERR_SURFACENOTATTACHED";
	case DDERR_TOOBIGHEIGHT:                 return "DDERR_TOOBIGHEIGHT";
	case DDERR_TOOBIGSIZE:                   return "DDERR_TOOBIGSIZE";
	case DDERR_TOOBIGWIDTH:                  return "DDERR_TOOBIGWIDTH";
	case DDERR_UNSUPPORTED:                  return "DDERR_UNSUPPORTED";
	case DDERR_UNSUPPORTEDFORMAT:            return "DDERR_UNSUPPORTEDFORMAT";
	case DDERR_UNSUPPORTEDMASK:              return "DDERR_UNSUPPORTEDMASK";
	case DDERR_VERTICALBLANKINPROGRESS:      return "DDERR_VERTICALBLANKINPROGRESS";
	case DDERR_WASSTILLDRAWING:              return "DDERR_WASSTILLDRAWING";
	case DDERR_XALIGN:                       return "DDERR_XALIGN";
	case DDERR_INVALIDDIRECTDRAWGUID:        return "DDERR_INVALIDDIRECTDRAWGUID";
	case DDERR_DIRECTDRAWALREADYCREATED:     return "DDERR_DIRECTDRAWALREADYCREATED";
	case DDERR_NODIRECTDRAWHW:               return "DDERR_NODIRECTDRAWHW";
	case DDERR_PRIMARYSURFACEALREADYEXISTS:  return "DDERR_PRIMARYSURFACEALREADYEXISTS";
	case DDERR_NOEMULATION:                  return "DDERR_NOEMULATION";
	case DDERR_REGIONTOOSMALL:               return "DDERR_REGIONTOOSMALL";
	case DDERR_CLIPPERISUSINGHWND:           return "DDERR_CLIPPERISUSINGHWND";
	case DDERR_NOCLIPPERATTACHED:            return "DDERR_NOCLIPPERATTACHED";
	case DDERR_NOHWND:                       return "DDERR_NOHWND";
	case DDERR_HWNDSUBCLASSED:               return "DDERR_HWNDSUBCLASSED";
	case DDERR_HWNDALREADYSET:               return "DDERR_HWNDALREADYSET";
	case DDERR_NOPALETTEATTACHED:            return "DDERR_NOPALETTEATTACHED";
	case DDERR_NOPALETTEHW:                  return "DDERR_NOPALETTEHW";
	case DDERR_BLTFASTCANTCLIP:              return "DDERR_BLTFASTCANTCLIP";
	case DDERR_NOBLTHW:                      return "DDERR_NOBLTHW";
	case DDERR_NODDROPSHW:                   return "DDERR_NODDROPSHW";
	case DDERR_OVERLAYNOTVISIBLE:            return "DDERR_OVERLAYNOTVISIBLE";
	case DDERR_NOOVERLAYDEST:                return "DDERR_NOOVERLAYDEST";
	case DDERR_INVALIDPOSITION:              return "DDERR_INVALIDPOSITION";
	case DDERR_NOTAOVERLAYSURFACE:           return "DDERR_NOTAOVERLAYSURFACE";
	case DDERR_EXCLUSIVEMODEALREADYSET:      return "DDERR_EXCLUSIVEMODEALREADYSET";
	case DDERR_NOTFLIPPABLE:                 return "DDERR_NOTFLIPPABLE";
	case DDERR_CANTDUPLICATE:                return "DDERR_CANTDUPLICATE";
	case DDERR_NOTLOCKED:                    return "DDERR_NOTLOCKED";
	case DDERR_CANTCREATEDC:                 return "DDERR_CANTCREATEDC";
	case DDERR_NODC:                         return "DDERR_NODC";
	case DDERR_WRONGMODE:                    return "DDERR_WRONGMODE";
	case DDERR_IMPLICITLYCREATED:            return "DDERR_IMPLICITLYCREATED";
	case DDERR_NOTPALETTIZED:                return "DDERR_NOTPALETTIZED";
	case DDERR_UNSUPPORTEDMODE:              return "DDERR_UNSUPPORTEDMODE";
	case DDERR_NOMIPMAPHW:                   return "DDERR_NOMIPMAPHW";
	case DDERR_INVALIDSURFACETYPE:           return "DDERR_INVALIDSURFACETYPE";
	case DDERR_DCALREADYCREATED:             return "DDERR_DCALREADYCREATED";
	case DDERR_CANTPAGELOCK:                 return "DDERR_CANTPAGELOCK";
	case DDERR_CANTPAGEUNLOCK:               return "DDERR_CANTPAGEUNLOCK";
	case DDERR_NOTPAGELOCKED:                return "DDERR_NOTPAGELOCKED";
	case DDERR_NOTINITIALIZED:               return "DDERR_NOTINITIALIZED";
	}
	return "Unknown Error";
}


// DXWindow-related libraries
#include "basicgfx.h"
#include "amorph_tilengine.h"
